unit UStepperMotorControl;

interface

uses
  Classes, SysUtils, StrUtils, CommPort, Math, ExtCtrls;

type
  TDriverCommand = (dcNoOperation, dcMotor1SetStepPeriod, dcMotor1SetFreeWheel,
    dcMotor1SetHalfStep, dcSetBaudRate, dcMotor1SetOutputLines, dcMotor2SetStepPeriod,
    dcMotor2SEtFreeWheel, dcMotor2SetHalfStep, dcMotor2SetOutputLines, dcMotorsEnable);
  TSingleMotorCommand = (smcMotorSetStepPeriod, smcMotorSetFreeWheel,
    smcMotorSetHalfStep, smcMotorSetOutputLines);

const
  MotorDriverCommands: array[0..1, TSingleMotorCommand] of TDriverCommand =
    ((dcMotor1SetStepPeriod, dcMotor1SetFreeWheel, dcMotor1SetHalfStep, dcMotor1SetOutputLines),
    (dcMotor2SetStepPeriod, dcMotor2SetFreeWheel, dcMotor2SetHalfStep, dcMotor2SetOutputLines));
  StepPeriodRefreshInterval = 20; // ms

type
  TMotorNumber = 0..1;

  TMotorControlProtocol = class
  private
    FStepPeriod: SmallInt;
    procedure TimerExecute(Sender: TObject);
    procedure RealSetMotorStepPeriod(Period: SmallInt);
  public
    Timer: TTimer;
    CommPort: TCommPort;
    MotorNumber: TMotorNumber;
    procedure Init;
    procedure NoOperation;
    procedure SetBaudRate(NewBaudRate: LongWord);
    procedure SetMotorStepPeriod(Period: SmallInt);
    procedure SetMotorFreeWheel(Enabled: Boolean);
    procedure SetMotorHalfStep(Enabled: Boolean);
    procedure SetMotorOutputLines(Enabled: Boolean);
    procedure SetMotorsEnable(Enabled: Boolean);
    constructor Create;
    destructor Destroy; override;
  end;

  TStepperMotor = class
  private
    FMotorControlProtocol: TMotorControlProtocol;
    FSpeedDivider: Integer;
    FStepPeriod: SmallInt;
    FHalfStep: Boolean;
    FFreeWheel: Boolean;
    FOutputLines: Boolean;
    FOnChange: TNotifyEvent;
    procedure SetStepPeriod(const Value: SmallInt);
    function GetRevolutionsPerSecond: Real;
    procedure SetRevolutionsPerSecond(const Value: Real);
    procedure SetStepsPerSeconds(const Value: Real);
    function GetStepsPerSeconds: Real;
    procedure SetFreeWheel(const Value: Boolean);
    function GetCommPort: TCommPort;
    procedure SetCommPort(const Value: TCommPort);
    function GetMotorNumber: TMotorNumber;
    procedure SetMotorNumber(const Value: TMotorNumber);
    procedure SetHalfStep(const Value: Boolean);
    procedure SetOutputLines(const Value: Boolean);
    function GetSpeed: Real;
    procedure SetSpeed(const Value: Real);
    procedure SetMotorsEnabled(const Value: Boolean);
  public
    MinRevolutions: Real;
    MaxRevolutions: Real;
    StepsPerRevolution: Integer;
    BasicPeriod: Real;
    WheelDiameter: Real;
    ReverseSpeed: Boolean;
    constructor Create;
    destructor Destroy; override;
    property MotorControlProtocol: TMotorControlProtocol read FMotorControlProtocol;
    property CommPort: TCommPort read GetCommPort write SetCommPort;
    property MotorNumber: TMotorNumber read GetMotorNumber write SetMotorNumber;
    property StepPeriod: SmallInt read FStepPeriod write SetStepPeriod;
    property FreeWheel: Boolean read FFreeWheel write SetFreeWheel;
    property HalfStep: Boolean read FHalfStep write SetHalfStep;
    property OutputLines: Boolean read FOutputLines write SetOutputLines;
    property RevolutionsPerSecond: Real read GetRevolutionsPerSecond write SetRevolutionsPerSecond;
    property StepsPerSeconds: Real read GetStepsPerSeconds write SetStepsPerSeconds;
    property Speed: Real read GetSpeed write SetSpeed;
    property MotorsEnabled: Boolean write SetMotorsEnabled;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  TDifferentialDrive = class
  private
    FWheelsDiameter: Real;
    FOnMove: TNotifyEvent;
    FOnChange: TNotifyEvent;
    FSpeed: Real;
    FAngle: Real;
    FY: Real;
    FX: Real;
    FOmega: Real;
    FTimer: TTimer;
    procedure TimerEvent(Sender: TObject);
    procedure SetWheelDiameter(const Value: Real);
    procedure SetSpeed(const Value: Real);
    procedure SetAllStop(const Value: Boolean);
    procedure SetOmega(const Value: Real);
    procedure ComputeMotorsSpeed;
    procedure ComputeAngleOmega;
    procedure ChangeEvent(Sender: TObject);
  public
    WheelsDistance: Real;
    Motors: array[TMotorNumber] of TStepperMotor;
    CommPort: TCommPort;
    constructor Create;
    procedure Reset;
    procedure MoveTo(X, Y: Real);   // not implemented
    procedure MoveRel(X, Y: Real);  // not implemented
    destructor Destroy; override;
    property X: Real read FX;
    property Y: Real read FY;
    property Angle: Real read FAngle;
    property Speed: Real read FSpeed write SetSpeed;
    property Omega: Real read FOmega write SetOmega;
    property WheelsDiameter: Real read FWheelsDiameter write SetWheelDiameter;
    property AllStop: Boolean write SetAllStop;
    property OnMove: TNotifyEvent read FOnMove write FOnMove;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

{ TStepperMotor }

function FloatMod(Value, Divider: Real): Real;
begin
  Result := Value - Trunc(Value / Divider) * Divider;
end;

constructor TStepperMotor.Create;
const
  PrescalerDivider = 64;
  CPUFrequency = 16000000; // ATmega16 16 MHz
begin
  FMotorControlProtocol := TMotorControlProtocol.Create;
  StepsPerRevolution := 200;
  BasicPeriod := 2 * PrescalerDivider / CPUFrequency;
  FSpeedDivider := 1;
  MaxRevolutions := 1;
  MinRevolutions := 0.1;
  ReverseSpeed := False;
end;

destructor TStepperMotor.Destroy;
begin
  FMotorControlProtocol.Free;
  inherited;
end;

function TStepperMotor.GetCommPort: TCommPort;
begin
  Result := FMotorControlProtocol.CommPort;
end;

function TStepperMotor.GetMotorNumber: TMotorNumber;
begin
  Result := FMotorControlProtocol.MotorNumber;
end;

function TStepperMotor.GetRevolutionsPerSecond: Real;
begin
  if FStepPeriod = 0 then Result := 0
    else Result := StepsPerSeconds / StepsPerRevolution;
end;

function TStepperMotor.GetSpeed: Real;
begin
  Result := RevolutionsPerSecond * Pi * WheelDiameter;
end;

function TStepperMotor.GetStepsPerSeconds: Real;
begin
  if FStepPeriod = 0 then Result := 0
    else Result := (1 / BasicPeriod) / FStepPeriod / FSpeedDivider;
  if ReverseSpeed then Result := -Result;
end;

procedure TStepperMotor.SetCommPort(const Value: TCommPort);
begin
  FMotorControlProtocol.CommPort := Value;
end;

procedure TStepperMotor.SetFreeWheel(const Value: Boolean);
begin
  FFreeWheel := Value;
  FMotorControlProtocol.SetMotorFreeWheel(Value);
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TStepperMotor.SetHalfStep(const Value: Boolean);
begin
  FHalfStep := Value;
  if Value then begin
    FSpeedDivider := 2;
    StepPeriod := FStepPeriod div 2;
  end else begin
    FSpeedDivider := 1;
    StepPeriod := FStepPeriod * 2;
  end;
  FMotorControlProtocol.SetMotorHalfStep(Value);
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TStepperMotor.SetMotorNumber(const Value: TMotorNumber);
begin
  FMotorControlProtocol.MotorNumber := Value;
end;

procedure TStepperMotor.SetMotorsEnabled(const Value: Boolean);
begin
  FMotorControlProtocol.SetMotorsEnable(Value);
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TStepperMotor.SetOutputLines(const Value: Boolean);
begin
  FOutputLines := Value;
  FMotorControlProtocol.SetMotorOutputLines(Value);
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TStepperMotor.SetRevolutionsPerSecond(const Value: Real);
begin
  if Value = 0 then StepPeriod := 0 else
    StepsPerSeconds := Value * StepsPerRevolution;
end;

procedure TStepperMotor.SetSpeed(const Value: Real);
begin
  RevolutionsPerSecond := Value / (Pi * WheelDiameter);
end;

procedure TStepperMotor.SetStepPeriod(const Value: SmallInt);
begin
  FMotorControlProtocol.SetMotorStepPeriod(Value);
  FStepPeriod := Value;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TStepperMotor.SetStepsPerSeconds(const Value: Real);
begin
  StepPeriod := Round((1 / BasicPeriod) / Value / FSpeedDivider);
  if ReverseSpeed then StepPeriod := -StepPeriod;
  if Assigned(FOnChange) then FOnChange(Self);
end;

{ TComProtocol }

constructor TMotorControlProtocol.Create;
begin
  FStepPeriod := 0;
  Timer := TTimer.Create(nil);
  Timer.Interval := StepPeriodRefreshInterval;
  Timer.OnTimer := TimerExecute;
  Timer.Enabled := True;
end;

destructor TMotorControlProtocol.Destroy;
begin
  Timer.Free;
  inherited;
end;

procedure TMotorControlProtocol.Init;
begin
  CommPort.PutChar(Char(dcNoOperation));
  CommPort.PutChar(Char(dcNoOperation));
  CommPort.PutChar(Char(dcNoOperation));
end;

procedure TMotorControlProtocol.NoOperation;
begin
  CommPort.PutChar(Char(dcNoOperation));
end;

procedure TMotorControlProtocol.RealSetMotorStepPeriod(Period: SmallInt);
begin
  CommPort.PutChar(Char(MotorDriverCommands[MotorNumber, smcMotorSetStepPeriod]));
  CommPort.PutChar(Char(Period and $ff));
  CommPort.PutChar(Char((Period shr 8) and $ff));
end;

procedure TMotorControlProtocol.SetBaudRate(NewBaudRate: LongWord);
begin
  with CommPort do begin
    PutChar(Char(dcSetBaudRate));
    PutChar(Char(NewBaudRate and $ff));
    PutChar(Char((NewBaudRate shr 8) and $ff));
    Sleep(1000);
    Open := False;
    Baud := NewBaudRate;
    Open := True;
  end;
end;

procedure TMotorControlProtocol.SetMotorFreeWheel(Enabled: Boolean);
begin
  CommPort.PutChar(Char(MotorDriverCommands[MotorNumber, smcMotorSetFreeWheel]));
  if Enabled then CommPort.PutChar(Char(0))
    else CommPort.PutChar(Char(1));
end;

procedure TMotorControlProtocol.SetMotorHalfStep(Enabled: Boolean);
begin
  CommPort.PutChar(Char(MotorDriverCommands[MotorNumber, smcMotorSetHalfStep]));
  if Enabled then CommPort.PutChar(Char(1))
    else CommPort.PutChar(Char(0));
end;

procedure TMotorControlProtocol.SetMotorOutputLines(Enabled: Boolean);
begin
  CommPort.PutChar(Char(MotorDriverCommands[MotorNumber, smcMotorSetOutputLines]));
  if Enabled then CommPort.PutChar(Char(1))
    else CommPort.PutChar(Char(0));
end;

procedure TMotorControlProtocol.SetMotorsEnable(Enabled: Boolean);
begin
  CommPort.PutChar(Char(dcMotorsEnable));
  if Enabled then CommPort.PutChar(Char(1))
    else CommPort.PutChar(Char(0));
end;

procedure TMotorControlProtocol.SetMotorStepPeriod(Period: SmallInt);
begin
  FStepPeriod := Period;
end;

{ TDiferentialDrive }

constructor TDifferentialDrive.Create;
begin
  CommPort := TCommPort.Create(nil);
  CommPort.ComNumber := 1;
  CommPort.Baud := 9600;
  CommPort.FlowControl := fcNone;
  CommPort.Open := True;
  Motors[0] := TStepperMotor.Create;
  Motors[0].CommPort := CommPort;
  Motors[0].MotorNumber := 0;
  Motors[0].OnChange := ChangeEvent;
  Motors[1] := TStepperMotor.Create;
  Motors[1].CommPort := CommPort;
  Motors[1].MotorNumber := 1;
  Motors[1].OnChange := ChangeEvent;
  Motors[1].ReverseSpeed := True;

  FX := 0;
  FY := 0;
  FAngle := 0;
  FSpeed := 0;
  FOmega := 0;
  FTimer := TTimer.Create(nil);
  FTimer.Interval := 100;
  FTimer.OnTimer := TimerEvent;
  FTimer.Enabled := True;
end;

destructor TDifferentialDrive.Destroy;
begin
  Motors[0].Free;
  Motors[1].Free;
  CommPort.Free;
  inherited;
end;

procedure TDifferentialDrive.MoveRel(X, Y: Real);
var
  Distance: Real;
begin
  Distance := Sqrt(Sqr(X) + Sqr(Y));
//  Angle := ArcTan2((Y - Self.Y), (X - Self.X)));
//  GoForward(Distance);
end;

procedure TDifferentialDrive.MoveTo(X, Y: Real);
var
  Distance: Real;
begin
  Distance := Sqrt(Sqr(X - Self.X) + Sqr(Y - Self.Y));
//  RotateTo(ArcTan2((Y - Self.Y), (X - Self.X)));
//  GoForward(Distance);
end;

procedure TDifferentialDrive.SetSpeed(const Value: Real);
begin
  FSpeed := Value;
  ComputeMotorsSpeed;
end;

procedure TDifferentialDrive.SetWheelDiameter(const Value: Real);
begin
  FWheelsDiameter := Value;
  Motors[0].WheelDiameter := Value;
  Motors[1].WheelDiameter := Value;
end;

procedure TDifferentialDrive.SetAllStop(const Value: Boolean);
begin
  Motors[0].MotorsEnabled := not Value;
end;

procedure TDifferentialDrive.SetOmega(const Value: Real);
begin
  FOmega := Value;
  ComputeMotorsSpeed;
end;

procedure TDifferentialDrive.ComputeMotorsSpeed;
begin
  Motors[0].OnChange := nil;
  Motors[1].OnChange := nil;
  Motors[0].Speed := (Speed - Omega * WheelsDistance) / 2;
  Motors[1].Speed := (Speed + Omega * WheelsDistance) / 2;
  Motors[0].OnChange := ChangeEvent;
  Motors[1].OnChange := ChangeEvent;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDifferentialDrive.TimerEvent(Sender: TObject);
begin
  FX := FX + Speed * Cos(Angle) * FTimer.Interval * 0.001;
  FY := FY + Speed * Sin(Angle) * FTimer.Interval * 0.001;
  FAngle := FloatMod(FAngle + Omega * FTimer.Interval * 0.001, 360);
  if Assigned(FOnMove) then FOnMove(Self);
end;

procedure TDifferentialDrive.Reset;
begin
  FX := 0;
  FY := 0;
  FAngle := 0;
  Omega := 0;
  Speed := 0;
end;

procedure TDifferentialDrive.ComputeAngleOmega;
begin
  FOmega := (Motors[1].Speed - Motors[0].Speed) / WheelsDistance;
  FSpeed := (Motors[1].Speed + Motors[0].Speed) / 2;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TDifferentialDrive.ChangeEvent(Sender: TObject);
begin
  ComputeAngleOmega;
end;

procedure TMotorControlProtocol.TimerExecute(Sender: TObject);
begin
  RealSetMotorStepPeriod(FStepPeriod);
end;

end.
